// Author: Matt Claessen, Avantes
// Date : 05-06-2008
//
// Minimal QT4 Linux application, for a single spectrometer channel
// This is the RS-232 version of the program. 
//
// Software used:
// Ubuntu 7.1 Desktop Edition
// QT 4.3.2 Open Source Edition (www.trolltech.com)
//
// The values for the detector are hardcoded for the Sony ILX 554
// Other detectors have a different number of dead pixels and a different number
// of datapixels
// You will have to change the measurement data structure (xxxxx_measdatatype)
// e.g. The Toshiba TCD-1304 has 13 dead pixels and 3648 datapixels
//      The Hamamatsu 9840 has 8 dead pixels and 1024 datapixels
//
// The low-level protocol used is described in the following Avantes document:
// "AS-5216 RS-232 communication protocol.pdf", available on request.
// This document describes the protocol that is used in both USB and RS-232
// communication. In this demo, only the most important commands are handled.
// Only the structures that are actually used have been changed to include a
// prefix and a postfix.
//
// Use a 3-wire null modem cable (no handshaking lines) with the following
// pinout to connect your AS-5216 to a COM port of your PC :
//
//     PC                      AS-5216
//  Female DB9 plug         Male HD26 Plug
//
//     2 (TX)  --------------- 19 (RX)
//     3 (RX)  --------------- 10 (TX)
//     5 (GND) ---------------  1 (GND)
//
// If you don't want to wire your own cable, the Avantes order# for this
// cable is IC-DB26/DB9-2
//
// In this demo, all 2048 pixels are used. This number can be reduced by changing
// the l_PrepareMeasData.m_StartPixel and l_PrepareMeasData.m_StopPixel values.
// The number of pixels in the xxxxx_measdatatype has to be reduced accordingly.
// The number of dead pixels is always the same and must not be changed.
// (Unless you change the detectortype, of course)

#include <stdio.h>
#include <iostream>
#include <cstdlib>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <QtGui>
#include <QApplication>
#include <QPushButton>
#include <QFile>
#include "qtdemo.h"
#include "renderarea.h"
#include "timestamp.h"

Qtdemo::Qtdemo(QWidget *parent)
	: QDialog(parent)
{
	OpenCommBtn = new QPushButton(tr("Open Communication"));
	SerialNumLabel = new QLabel(tr("Serial Number"));
	SerialNum = new QLabel();
	CloseCommBtn = new QPushButton(tr("Close Communication"));
	StartMeasBtn = new QPushButton(tr("Start Measurement"));
	StopMeasBtn = new QPushButton(tr("Stop Measurement"));
	TimeStampLabel = new QLabel(tr("Timestamp:"));
	timeStamp = new TimeStamp;
	IntTimeLabel = new QLabel(tr("Integration Time [ms]:"));
	edtIntTime = new QLineEdit("5.0");
	NumAvgLabel = new QLabel(tr("Number of Averages:"));
	edtNumAvg = new QLineEdit("1");
	NumMeasLabel = new QLabel(tr("Number of Measurements:"));;
	edtNumMeas = new QLineEdit("1");

	QVBoxLayout *leftLayout = new QVBoxLayout;
	leftLayout->addWidget(OpenCommBtn);
	leftLayout->addWidget(SerialNumLabel);
	leftLayout->addWidget(SerialNum);
	leftLayout->addWidget(CloseCommBtn);
	leftLayout->addWidget(StartMeasBtn);
	leftLayout->addWidget(StopMeasBtn);
	leftLayout->addWidget(TimeStampLabel);
	leftLayout->addWidget(timeStamp);
	leftLayout->addWidget(IntTimeLabel);
	leftLayout->addWidget(edtIntTime);
	leftLayout->addWidget(NumAvgLabel);
	leftLayout->addWidget(edtNumAvg);
	leftLayout->addWidget(NumMeasLabel);
	leftLayout->addWidget(edtNumMeas);
	leftLayout->addStretch();
	renderArea = new RenderArea;
	renderArea->setPen(QPen(Qt::red, 0, Qt::SolidLine, Qt::SquareCap, Qt::BevelJoin));
	QGridLayout *mainLayout = new QGridLayout; 
	mainLayout->addLayout(leftLayout,0,0,0);
	mainLayout->addWidget(renderArea,0,1,0);
	mainLayout->setRowStretch(2,10);
	mainLayout->setColumnStretch(3,10); 
	setLayout(mainLayout);
	connect(OpenCommBtn, SIGNAL(clicked()), this, SLOT(on_OpenCommBtn_clicked()));
	connect(CloseCommBtn, SIGNAL(clicked()), this, SLOT(on_CloseCommBtn_clicked()));
	connect(StartMeasBtn, SIGNAL(clicked()), this, SLOT(on_StartMeasBtn_clicked()));
	connect(StopMeasBtn, SIGNAL(clicked()), this, SLOT(on_StopMeasBtn_clicked()));
}

Qtdemo::~Qtdemo()
{

}

// Please note the swap functions that reverse the byte order of all data,
// both incoming and outgoing. The AS-5216 uses a ColdFire microprocessor,
// which is of Motorola heritage and therefore big-endian. PC's are of Intel
// heritage and therefore little-endian.
// This means that the byte order of all 16-bit and 32-bit values has to be
// reversed.

float swapsingle(float floatin)
{
	union s
	{
		char sa[4];
		float res;
	} temp;
	temp.res = floatin;
	s floatout;
	for (int teller=0; teller<4; ++teller){
		floatout.sa[teller]=temp.sa[3-teller];
	}
    return floatout.res;
}

uint32 swap32(uint32 uint32in)
{
	union s
	{
		char sa[4];
		uint32 res;
	} temp;
	temp.res = uint32in;
	s uint32out;
	for (int teller=0; teller<4; ++teller){
		uint32out.sa[teller]=temp.sa[3-teller];
	}
    return uint32out.res;
}

uint16 swap16(uint16 uint16in)
{
	union s
	{
		char sa[2];
		uint16 res;
	} temp;
	temp.res = uint16in;
	s uint16out;
	for (int teller=0; teller<2; ++teller){
		uint16out.sa[teller]=temp.sa[1-teller];
	}
    return uint16out.res;
}

bool Qtdemo::sendarray(outbuftype outarray, int size)
{
	bool success=true;
	for (int teller=0;teller<size;teller++) {
		char * outchar;
		outchar = (char *) outarray+teller;
		if (write(fd, outchar, 1) < 0)
			success = false;
		if ((teller >= 2) &&					// past DLE/STX
			(teller <= (size-3)) && 			// before DLE/ETX
			(outarray[teller] == DLE))
				write(fd, &DLE, 1);				// duplicate DLE
	}
	return success;
}

void Qtdemo::on_OpenCommBtn_clicked()
{
	unsigned char bufout[12];
	unsigned char bufout2[12];

	if (!(fd = open("/dev/ttyS0",O_RDWR | O_NOCTTY | O_NDELAY)))  // Open COM1
		QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Could not open COM1"),
			QMessageBox::Ok, QMessageBox::Ok);
	fcntl(fd, F_SETFL, 0);		// reset all statusflags
	struct termios options;
	tcgetattr(fd, &options);		// set baud rate to 115200 bps
	cfsetispeed(&options, B115200);
	cfsetospeed(&options, B115200);
	options.c_cflag &= ~PARENB;		// parity N
	options.c_cflag &= ~CSTOPB;		// 1 stopbit
	options.c_cflag &= ~CSIZE;		// 8 databits
	options.c_cflag |= CS8;
	options.c_cflag |= (CLOCAL | CREAD);
	options.c_lflag &= ~(ICANON | ECHO | ECHOE |ISIG);  // Raw input
	options.c_iflag &= ~(IXON | IXOFF | IXANY); // disable software flow control
	options.c_oflag &= ~OPOST; // Raw output
	options.c_cc[VMIN] = 0;
	options.c_cc[VTIME] = 100; // 10 second time_out
	tcsetattr(fd, TCSANOW, &options);
	
	tcflush(fd, TCIOFLUSH);  // flush input & output buffers
	bufout[0] = DLE;
	bufout[1] = STX;
	bufout[2] = 0x20;
	bufout[3] = 0x00;
	bufout[4] = 0x01;  // node address
	bufout[5] = 0x00;  // reserved
	bufout[6] = 0x02;  // length of the command
	bufout[7] = 0x00;
	bufout[8] = 0x13;  // get_ident
	bufout[9] = 0x00;
	bufout[10] = DLE;
	bufout[11] = ETX;
	
	if (sendarray(bufout, sizeof(bufout)) == false)
		QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Writing to COM1 failed"),
			QMessageBox::Ok, QMessageBox::Ok);
	usleep(10000);
//	int bytes;
//	ioctl(fd, FIONREAD, &bytes);			// read size of input buffer	
	
	AvsIdentityType Avs_Id;
	if (read(fd,(char*) &Avs_Id,sizeof(Avs_Id)) < 0) 
		QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Reading from COM1 failed"),
			QMessageBox::Ok, QMessageBox::Ok);
	else SerialNum->setText(QString(Avs_Id.SerialNumber).left(9));
	
	tcflush(fd, TCIOFLUSH);  // flush input & output buffers
	bufout2[0] = DLE;
	bufout2[1] = STX;
	bufout2[2] = 0x20;
	bufout2[3] = 0x00;
	bufout2[4] = 0x01;  // node address
	bufout2[5] = 0x00;  // reserved
	bufout2[6] = 0x02;  // length of the command
	bufout2[7] = 0x00;
	bufout2[8] = 0x01;  // get_device_configuration
	bufout2[9] = 0x00;
	bufout2[10] = DLE;
	bufout2[11] = ETX;
	
	if ((sendarray(bufout2, sizeof(bufout2)) == false))
		QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Writing to COM1 failed"),
			QMessageBox::Ok, QMessageBox::Ok);
	usleep(10000);

//	ioctl(fd, FIONREAD, &bytes);		// read size of input buffer	
	DeviceConfigType devcon;	
	
	int thispass;
	unsigned int numread = 0;
	while (numread < sizeof(devcon)) {
		thispass = read(fd,(char*) &devcon+numread,sizeof(devcon)-numread);
		if ((thispass < 0) && (errno != EAGAIN)) {
			QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Reading from COM1 failed"),
			QMessageBox::Ok, QMessageBox::Ok);
		} else
		numread += thispass;
	}
   	for (int teller=0;teller<NR_WAVELEN_POL_COEF;++teller) {
		fit[teller]=swapsingle(devcon.m_Detector.m_aFit[teller]);
	
		extern unsigned short m_NrPixels;
		// This is the polynomial that is used to calculate wavelength values
		// from pixel values. The coefficients are stored in the eeprom of the
		// AS-5216 and extracted with the get_device_configuration command (above)
		extern double m_pLambda[MAX_NR_PIXELS];
		m_NrPixels=swap16(devcon.m_Detector.m_NrPixels);
		for (int teller=0;teller<m_NrPixels;++teller) {
			m_pLambda[teller]=	fit[0] +
								fit[1]*teller*1.0 +
								fit[2]*teller*teller*1.0 +
								fit[3]*teller*teller*teller*1.0 +
								fit[4]*teller*teller*teller*teller*1.0;
		}
	}
//	check intercept for debugging purposes	
//	QMessageBox::information(this, tr("QT4 Demo"),
//			tr("Intercept: %1").arg(fit[0]),
//			QMessageBox::Ok, QMessageBox::Ok);
}

void Qtdemo::on_CloseCommBtn_clicked()
{
	::close(fd);   // use global namespace, prototype is located in unistd.h
	SerialNum->setText("");
}

void Qtdemo::on_StartMeasBtn_clicked()
{
	unsigned char bufout2[14];
	unsigned char bufout3[12];
	unsigned int bufpoint = 0;
	uint32 numavg;
	bool ok;
	SendMeasConfigType l_PrepareMeasData;
	sony_single_measdatatype sony_single_meas;
	sony_multi_measdatatype sony_multi_meas;
	short l_NrOfScans = QString(edtNumMeas->text()).toInt(&ok);

	tcflush(fd, TCIOFLUSH);  // flush input & output buffers
	memset(&bufin, 0, sizeof(bufin));  // fill bufin with zeros
	l_PrepareMeasData.prefix[0] = DLE;
	l_PrepareMeasData.prefix[1] = STX;
	l_PrepareMeasData.prefix[2] = 0x20;
	l_PrepareMeasData.prefix[3] = 0x00;
	l_PrepareMeasData.prefix[4] = 0x01;   // node address
	l_PrepareMeasData.prefix[5] = 0x00;   // reserved
	l_PrepareMeasData.prefix[6] = 0x2B;   // length of the command
	l_PrepareMeasData.prefix[7] = 0x00;
	l_PrepareMeasData.prefix[8] = 0x05;   // prepare_measurement
	l_PrepareMeasData.prefix[9] = 0x00;
	l_PrepareMeasData.m_Meas.m_StartPixel = 0;
	extern unsigned short m_NrPixels;
	l_PrepareMeasData.m_Meas.m_StopPixel = swap16(m_NrPixels-1);
	QLocale::setDefault(QLocale::Dutch);
	double inttime = QString(edtIntTime->text()).toDouble(&ok);
	l_PrepareMeasData.m_Meas.m_IntegrationTime = swapsingle(inttime);
	l_PrepareMeasData.m_Meas.m_IntegrationDelay = 0;
	numavg =QString(edtNumAvg->text()).toInt(&ok);
	l_PrepareMeasData.m_Meas.m_NrAverages = swap32(numavg);
	l_PrepareMeasData.m_Meas.m_CorDynDark.m_Enable = 0;
	l_PrepareMeasData.m_Meas.m_CorDynDark.m_ForgetPercentage = 0;
	l_PrepareMeasData.m_Meas.m_Smoothing.m_SmoothPix = 0;
	l_PrepareMeasData.m_Meas.m_Smoothing.m_SmoothModel = 0;
	l_PrepareMeasData.m_Meas.m_SaturationDetection = 0;
	l_PrepareMeasData.m_Meas.m_Trigger.m_Mode = 0;
	l_PrepareMeasData.m_Meas.m_Trigger.m_Source = 0;
	l_PrepareMeasData.m_Meas.m_Trigger.m_SourceType = 0;
	l_PrepareMeasData.m_Meas.m_Control.m_StrobeControl = 0;
	l_PrepareMeasData.m_Meas.m_Control.m_LaserDelay = 0;
	l_PrepareMeasData.m_Meas.m_Control.m_LaserWidth = 0;
	l_PrepareMeasData.m_Meas.m_Control.m_LaserWaveLength = 0;
	l_PrepareMeasData.m_Meas.m_Control.m_StoreToRam = 0;
	l_PrepareMeasData.postfix[0] = DLE;
	l_PrepareMeasData.postfix[1] = ETX;
	
	memcpy(&outarray,&l_PrepareMeasData,sizeof(l_PrepareMeasData));
	if (sendarray(outarray, sizeof(l_PrepareMeasData)) == false)
			QMessageBox::critical(this, tr("QT4 Demo"),
				tr("Writing to COM1 failed"),
				QMessageBox::Ok, QMessageBox::Ok);
	usleep(10000);
	int retval;
	retval = read(fd,(char*) &bufin, 12);			// read 12 byte response
	if (retval < 0)									// to prepare_measurement
		QMessageBox::critical(this, tr("QT4 Demo"),		
			tr("Reading from COM1 failed"),
			QMessageBox::Ok, QMessageBox::Ok);
	else {
		if ((retval != 12 ) || (bufin[8] != 0x85))
			QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Error in Prepare_Measurement"),
			QMessageBox::Ok, QMessageBox::Ok);
	}
	
	tcflush(fd, TCIOFLUSH);  // flush input & output buffers
	memset(&bufin, 0, sizeof(bufin));  // fill bufin with zeros
	bufpoint = 0;
	bufout2[0] = DLE;
	bufout2[1] = STX;
	bufout2[2] = 0x20;
	bufout2[3] = 0x00;
	bufout2[4] = 0x01;   // node address
	bufout2[5] = 0x00;   // reserved
	bufout2[6] = 0x04;   // length of the command
	bufout2[7] = 0x00;
	bufout2[8] = 0x06;   // start_measurement
	bufout2[9] = 0x00;
	bufout2[10] = HIBYTE(l_NrOfScans);
	bufout2[11] = LOBYTE(l_NrOfScans);
	bufout2[12] = DLE;
	bufout2[13] = ETX;

	if ((sendarray(bufout2, sizeof(bufout2)) == false))
		QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Writing to COM1 failed"),
			QMessageBox::Ok, QMessageBox::Ok);
	usleep(10000);
	retval = read(fd,(char*) &bufin, 12);				// read 12 byte response
	if (retval <  0)									// to start_measurement
		QMessageBox::critical(this, tr("QT4 Demo"),			
			tr("Reading from COM1 failed"),
			QMessageBox::Ok, QMessageBox::Ok);
	else {
		if ((retval != 12 ) || (bufin[8] != 0x86))
			QMessageBox::critical(this, tr("QT4 Demo"),
			tr("Error in Start_Measurement"),
			QMessageBox::Ok, QMessageBox::Ok);
		else bufpoint += 12;
	}
	
	// bufin will now contain 2 frames, therefore increment the buffer pointer with 12
	
	extern unsigned int l_Time;
	extern double l_pSpectrum[MAX_NR_PIXELS];
	unsigned int request_size;
	if (numavg <= 1)
		request_size = sizeof(sony_single_meas);	// 4148 bytes
	else
		request_size = sizeof(sony_multi_meas);		// 8282 bytes
	short measnr = 0;
	while (measnr < l_NrOfScans) {
		int thispass;
 	  	unsigned int numread = 0;
		while (numread < request_size) {
			thispass = read(fd,(char*) &bufin + bufpoint + numread, 
							request_size - numread);
			if ((thispass < 0) && (errno != EAGAIN)) {
				QMessageBox::critical(this, tr("QT4 Demo"),
				tr("Reading from COM1 failed"),
				QMessageBox::Ok, QMessageBox::Ok);
			} else
			numread += thispass;
		}
		// read the rest until DLE/ETX marker (there may be more due to doubling of DLE's in data)
		bool endofframe = false;
		unsigned char byte1 = 0;
		unsigned char byte2 = bufin[numread + bufpoint - 1];  // to do -> check start value
		
//		QMessageBox::information(this, tr("QT4 Demo"),
//		tr("numread: %1 byte2: %2").arg(numread) .arg(byte2),
//		QMessageBox::Ok, QMessageBox::Ok);
		
		while (endofframe == false) {
			byte1 = byte2;
			read(fd, &byte2, 1);
			bufin[numread + bufpoint] = byte2;
			numread++;
			if ((byte1 == DLE) && (byte2 == ETX))
				endofframe = true;
		}
		
//		dump bufin to a file for debugging
//		QFile file1("bufin1.dat");
//		if (!file1.open(QIODevice::WriteOnly)) {
//			std::cerr << "Cannot open file for writing: "
//				<< qPrintable(file1.errorString()) << std::endl;
//		}
//		QDataStream out1(&file1);
//		for (unsigned int i=0;i<numread+bufpoint;i++)
//			out1 << bufin[i];
		
//		QMessageBox::information(this, tr("Before filterDLE"),
//		tr("numread: %1 selector: %2").arg(numread) .arg(bufin[bufpoint + 8] ),
//		QMessageBox::Ok, QMessageBox::Ok);
		
		byte1 = 0;
		byte2 = 0;
		unsigned int readcount = 0;
		unsigned int writecount = 0;
		inbuftype tempbuf;
		memset(&tempbuf, 0, sizeof(tempbuf));  // fill tempbuf with zeros
		while (readcount < (numread+bufpoint)) {
			byte1=byte2;
			byte2=bufin[readcount];
			if ((byte1 == DLE) && (byte2 == DLE)) {
				readcount++;
				byte2 = 0;	// filter DLE, clear byte2 to prevent filtering too much
			}				// e.g. 4 consecutive DLE's should be filtered to 2 DLE's, 
							// and not just to 1 DLE
			else {  // this is the normal case
				tempbuf[writecount]=byte2;
				readcount++;
				writecount++;
			}
		}
		memcpy(bufin, tempbuf, writecount);	// copy filtered data back to bufin
		numread = writecount - bufpoint;	// restore expected value
		
//		QMessageBox::information(this, tr("After filterDLE"),
//		tr("numread: %1 selector: %2").arg(numread) .arg(bufin[bufpoint + 8] ),
//		QMessageBox::Ok, QMessageBox::Ok);

//		// dump bufin to a file for debugging
//		QFile file2("bufin2.dat");
//		if (!file2.open(QIODevice::WriteOnly)) {
//			std::cerr << "Cannot open file for writing: "
//				<< qPrintable(file2.errorString()) << std::endl;
//		}
//		QDataStream out2(&file2);
//		for (unsigned int i=0;i<numread+bufpoint;i++)  
//			out2 << bufin[i];
		
		if (numread >= request_size) {
			switch (bufin[bufpoint + 8])
			{
				case 0xB0:	// measurement_data with #averages = 1
					memcpy(&sony_single_meas,bufin+bufpoint,sizeof(sony_single_meas));
					l_Time = swap32(sony_single_meas.timestamp);
					for (int teller=0;teller<m_NrPixels;++teller) 
						l_pSpectrum[teller]=swap16(sony_single_meas.pixels[teller]); 
					break;
				case 0xB1:	// measurement_data with #averages > 1
					memcpy(&sony_multi_meas,bufin+bufpoint,sizeof(sony_multi_meas));
					l_Time = swap32(sony_multi_meas.timestamp);
					if (numavg != swap16(sony_multi_meas.averages)) 
						QMessageBox::critical(this, tr("QT4 Demo"),
							tr("Error in Number of Averages"),
							QMessageBox::Ok, QMessageBox::Ok);
					for (int teller=0;teller<m_NrPixels;++teller) 
						l_pSpectrum[teller]=swap32(sony_multi_meas.pixels[teller])/numavg; 
					break;
				default:	// should not occur
					QMessageBox::critical(this, tr("QT4 Demo"),
							tr("Error in selector byte"),
							QMessageBox::Ok, QMessageBox::Ok);
					break;	
			}
		}
		++measnr;
		renderArea->repaint();
		timeStamp->repaint();
		// QT appears to write only 1 in 2 spectra to the screen
		// if you save the spectra to disk, you will see that the full number is available 
		
// 		dump l_pSpectrum to a file for debugging
//		QFile file3(QString("%1").arg(l_Time));
//		if (!file3.open(QIODevice::WriteOnly)) {
//			std::cerr << "Cannot open file for writing: "
//				<< qPrintable(file3.errorString()) << std::endl;
//		}
//		QDataStream out3(&file3);
//		for (unsigned int i=0;i<m_NrPixels;i++)  
//			out3 << l_pSpectrum[i];
	
		if (measnr<l_NrOfScans) {
			tcflush(fd, TCIOFLUSH);  // flush input & output buffers
			memset(&bufin, 0, sizeof(bufin));  // fill bufin with zeros
			bufpoint = 0;
			bufout3[0] = DLE;
			bufout3[1] = STX;
			bufout3[2] = 0x21;
			bufout3[3] = 0x00;
			bufout3[4] = 0x01;  // node address
			bufout3[5] = 0x00;  // reserved 
			bufout3[6] = 0x02;  // length of the command
			bufout3[7] = 0x00;
			bufout3[8] = 0xC0;  // acknowledge
			bufout3[9] = 0x00;
			bufout3[10] = DLE;
			bufout3[11] = ETX;
			if ((sendarray(bufout3, sizeof(bufout3)) == false))
				QMessageBox::critical(this, tr("QT4 Demo"),
				tr("Writing to COM1 failed"),
				QMessageBox::Ok, QMessageBox::Ok);
		}
	} // while
}

void Qtdemo::on_StopMeasBtn_clicked()
{
	// nothing for now
}

